/* -----------------------------------------------------------------------------
*
*  npmodels.h
*
*  ANTz - realtime 3D data visualization tools for the real-world, based on NPE.
*
*  ANTz is hosted at http://openantz.com and NPE at http://neuralphysics.org
*
*  Written in 2010-2016 by Shane Saxon - saxon@openantz.com
*
*  Please see main.c for a complete list of additional code contributors.
*
*  To the extent possible under law, the author(s) have dedicated all copyright 
*  and related and neighboring rights to this software to the public domain
*  worldwide. This software is distributed without any warranty.
*
*  Released under the CC0 license, which is GPL compatible.
*
*  You should have received a copy of the CC0 Public Domain Dedication along
*  with this software (license file named LICENSE.txt). If not, see
*  http://creativecommons.org/publicdomain/zero/1.0/
*
* --------------------------------------------------------------------------- */

#ifndef NPMODELS_H_
#define NPMODELS_H_

#include "../../npdata.h"

/// List of models table keeps track of file name, path, scale, orientation,
/// polygon count, GPU memory required, etc.
struct NPmodels {
	int		id;		///< this models ID
	int		geometryID;		///< geomtery ID assigned for the mesh
	int		textureID;		///< texture map ID

	char	fileName[kNPmaxName];
	char	fileDir[kNPmaxPath];

	long long int gpuSize;		///< GPU memory required for the model geometry	
};							
typedef struct NPmodels NPmodels;
typedef struct NPmodels *pNPmodels;

/// Contains the file name, geometry_id and texture_id of the model
/// Separate 'models' table keeps track of additional attributes
/*
struct NPgeo {
	int		id;				///< geomtery ID assigned for the mesh
	int		textureID;		///< assigned texture ID or 0 for default
	int		modelsID;		///< models_id for getting additional attributes
};							
typedef struct NPgeo NPgeo;
typedef struct NPgeo *pNPgeo;
*/

//------------------------------------------------------------------------------
void npInitModels (void* dataRef);
void npCloseModels (void* dataRef);
//pNPgeolist npAddGeo(int* geoId, int* extTexId, int type, pNPfloatXYZ center, pNPfloatXYZ rotate, pNPfloatXYZ scale, char* object_name, char* file_name, char* path, void* dataRef);

//bool npGeoMapIDis(int extGeoId, void* dataRef);
pNPgeolist npGeoMapIDis(int extGeoId, void* dataRef);
int npGeoMapIDint(int extGeoId, void* dataRef);
int npGeoMapIDextTexID(int extGeoId, void* dataRef);
int npGeoMapIDisLoaded(int extGeoId, void* dataRef);
int npGeoMapIDintTexID(int intGeoId, void* dataRef);
int npLoadModel(pNPgeolist geo, void* dataRef);
pNPgeolist npLoadModelFromFile(char* filepath, void* dataRef);

void npUpdateGeoList( void* dataRef );

const struct aiScene* npModelImport(char* filePath, void* dataRef);

void npGeoSync (void* dataRef);

void npTexMapSync(void* dataRef);
void npTexMapSync2(void* dataRef);

pNPgeo npModelNew(char* model_csvline, void* dataRef);

pNPgeo npModelNew2(char* model_csvline, void* dataRef);

void npInitTmpGeoMap(void* dataRef);

char* npModelNewTypeId(char* csv_typeId, int* typeId, void* dataRef);

char* npModelNewObjectName(char* stringVal, int maxSize, char* objectName, void* dataRef);

char* npModelNewFileName(char* stringVal, int maxSize, char* fileName, int i, void* dataRef);

char* npModelNewFilePath(char* stringVal, int maxSize, char* filePath, int i, void* dataRef);



void npLoadExtGeos(void* dataRef);

#define kNPgetGeoId 0
#define kNPsetGeoId 1
#define kNPgeoLoadTex 2
#define kNPgeoLoad 3
#define kNPunsetGeoId 4
#define kNPgeoMapIntTexId 5
#define kNPgeoMapExtTexId 6
#define kNPgeoMapIDis 7
#define kNPgeoSetTex 8
#define kNPgetGeoIdB 9
void* npGeo(int action, int* xGeoId, int* texId, char* path, char* file, void* dataRef);

void npLoadGeos(void* dataRef);

#endif

